@@ -125,7 +125,9 @@ gupnp_service_get_session (GUPnPService *service)
                 if (g_getenv ("GUPNP_DEBUG")) {
                         SoupLogger *logger;
                         logger = soup_logger_new (SOUP_LOGGER_LOG_BODY, -1);
-                        soup_logger_attach (logger, service->priv->session);
+                        soup_session_add_feature (
+                                        service->priv->session,
+                                        SOUP_SESSION_FEATURE (logger));
                 }
         }
 
@@ -164,7 +166,7 @@ subscription_data_free (SubscriptionData *data)
         g_free (data->sid);
 
         if (data->timeout_src)
-	        g_source_destroy (data->timeout_src);
+                g_source_destroy (data->timeout_src);
 
         g_slice_free (SubscriptionData, data);
 }
@@ -191,6 +193,7 @@ struct _GUPnPServiceAction {
         char         *name;
 
         SoupMessage  *msg;
+        gboolean      accept_gzip;
 
         GUPnPXMLDoc  *doc;
         xmlNode      *node;
@@ -250,7 +253,6 @@ static void
 finalize_action (GUPnPServiceAction *action)
 {
         SoupServer *server;
-        char *response_body;
 
         /* Embed action->response_str in a SOAP document */
         g_string_prepend (action->response_str,
@@ -271,13 +273,22 @@ finalize_action (GUPnPServiceAction *action)
                          "</s:Body>"
                          "</s:Envelope>");
 
-        response_body = g_string_free (action->response_str, FALSE);
+        soup_message_headers_replace (action->msg->response_headers,
+                                      "Content-Type",
+                                      "text/xml; charset=\"utf-8\"");
 
-        soup_message_set_response (action->msg,
-                                   "text/xml; charset=\"utf-8\"",
-                                   SOUP_MEMORY_TAKE,
-                                   response_body,
-                                   strlen (response_body));
+        if (action->accept_gzip && action->response_str->len > 1024) {
+                http_response_set_body_gzip (action->msg,
+                                             action->response_str->str,
+                                             action->response_str->len);
+                g_string_free (action->response_str, TRUE);
+        } else {
+                soup_message_body_append (action->msg->response_body,
+                                          SOUP_MEMORY_TAKE,
+                                          action->response_str->str,
+                                          action->response_str->len);
+                g_string_free (action->response_str, FALSE);
+        }
 
         /* Server header on response */
         soup_message_headers_append
@@ -871,6 +882,7 @@ control_server_handler (SoupServer        *server,
         xmlDoc *doc;
         xmlNode *action_node;
         const char *soap_action;
+        const char *accept_encoding;
         char *action_name;
         char *end;
         GUPnPServiceAction *action;
@@ -892,8 +904,8 @@ control_server_handler (SoupServer        *server,
         context = gupnp_service_info_get_context (GUPNP_SERVICE_INFO (service));
 
         /* Get action name */
-        soap_action = soup_message_headers_get (msg->request_headers,
-                                                "SOAPAction");
+        soap_action = soup_message_headers_get_one (msg->request_headers,
+                                                    "SOAPAction");
         if (!soap_action) {
                 soup_message_set_status (msg, SOUP_STATUS_PRECONDITION_FAILED);
                 return;
@@ -936,7 +948,7 @@ control_server_handler (SoupServer        *server,
         }
 
         /* Create action structure */
-        action = g_slice_new (GUPnPServiceAction);
+        action = g_slice_new0 (GUPnPServiceAction);
 
         action->ref_count    = 1;
         action->name         = g_strdup (action_name);
@@ -947,6 +959,24 @@ control_server_handler (SoupServer        *server,
                                                         soap_action);
         action->context      = g_object_ref (context);
 
+        /* Get accepted encodings */
+        accept_encoding = soup_message_headers_get_list (msg->request_headers,
+                                                         "Accept-Encoding");
+
+        if (accept_encoding) {
+                GSList *codings;
+
+                codings = soup_header_parse_quality_list (accept_encoding,
+                                                          NULL);
+                if (codings &&
+                    g_slist_find_custom (codings, "gzip",
+                                         (GCompareFunc) g_ascii_strcasecmp)) {
+                       action->accept_gzip = TRUE;
+                }
+
+                soup_header_free_list (codings);
+        }
+
         /* Tell soup server that response is not ready yet */
         soup_server_pause_message (server, msg);
 
@@ -1128,8 +1158,8 @@ subscribe (GUPnPService *service,
         data->sid     = generate_sid ();
 
         /* Add timeout */
-	data->timeout_src = g_timeout_source_new_seconds (SUBSCRIPTION_TIMEOUT);
-	g_source_set_callback (data->timeout_src,
+        data->timeout_src = g_timeout_source_new_seconds (SUBSCRIPTION_TIMEOUT);
+        g_source_set_callback (data->timeout_src,
                                subscription_timeout,
                                data,
                                NULL);
@@ -1216,9 +1246,10 @@ subscription_server_handler (SoupServer        *server,
 
         service = GUPNP_SERVICE (user_data);
 
-        callback = soup_message_headers_get (msg->request_headers, "Callback");
-        nt       = soup_message_headers_get (msg->request_headers, "NT");
-        sid      = soup_message_headers_get (msg->request_headers, "SID");
+        callback = soup_message_headers_get_one (msg->request_headers,
+                                                 "Callback");
+        nt       = soup_message_headers_get_one (msg->request_headers, "NT");
+        sid      = soup_message_headers_get_one (msg->request_headers, "SID");
 
         /* Choose appropriate handler */
         if (strcmp (msg->method, GENA_METHOD_SUBSCRIBE) == 0) {
@@ -1281,7 +1312,7 @@ got_introspection (GUPnPServiceInfo          *info,
                    const GError              *error,
                    gpointer                   user_data)
 {
-        GUPnPService *service = user_data;
+        GUPnPService *service;
         const GList *state_variables, *l;
         GHashTableIter iter;
         gpointer data;
@@ -1340,7 +1371,6 @@ gupnp_service_constructor (GType                  type,
 {
         GObjectClass *object_class;
         GObject *object;
-        GUPnPService *service;
         GUPnPServiceInfo *info;
         GUPnPContext *context;
         SoupServer *server;
@@ -1354,7 +1384,6 @@ gupnp_service_constructor (GType                  type,
                                             n_construct_params,
                                             construct_params);
 
-        service = GUPNP_SERVICE (object);
         info    = GUPNP_SERVICE_INFO (object);
 
         /* Get introspection and save state variable names */
@@ -1386,14 +1415,6 @@ gupnp_service_constructor (GType                  type,
         return object;
 }
 
-static gboolean
-say_yes (gpointer key,
-         gpointer value,
-         gpointer user_data)
-{
-        return TRUE;
-}
-
 /* Root device availability changed. */
 static void
 notify_available_cb (GObject *object,
@@ -1406,9 +1427,7 @@ notify_available_cb (GObject *object,
 
         if (!gupnp_root_device_get_available (GUPNP_ROOT_DEVICE (object))) {
                 /* Root device now unavailable: Purge subscriptions */
-                g_hash_table_foreach_remove (service->priv->subscriptions,
-                                             say_yes,
-                                             NULL);
+                g_hash_table_remove_all (service->priv->subscriptions);
         }
 }
 
@@ -1567,7 +1586,6 @@ static void
 gupnp_service_class_init (GUPnPServiceClass *klass)
 {
         GObjectClass *object_class;
-        GUPnPServiceInfoClass *info_class;
 
         object_class = G_OBJECT_CLASS (klass);
 
@@ -1577,8 +1595,6 @@ gupnp_service_class_init (GUPnPServiceClass *klass)
         object_class->dispose      = gupnp_service_dispose;
         object_class->finalize     = gupnp_service_finalize;
 
-        info_class = GUPNP_SERVICE_INFO_CLASS (klass);
-
         g_type_class_add_private (klass, sizeof (GUPnPServicePrivate));
 
         /**
